import pytest

from src.py_cdll import NodesNotInSameListError
from src.py_cdll.node import Node, nodes_from_values, split, equal, is_consistent


def test_split_one_node_one_head_one_list_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values0)

    # Execution
    split(first_head=node0, second_head=node0)

    # Validation
    assert equal(first=node0, second=node1)
    assert is_consistent(node=node0) is True


def test_split_two_nodes_one_head_one_list_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values0)

    # Execution
    split(first_head=node0, second_head=node0)

    # Validation
    assert equal(first=node0, second=node1)
    assert is_consistent(node=node0) is True


def test_split_two_nodes_two_heads_one_list_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    values1: list[str] = ["zero"]
    values2: list[str] = ["one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next
    node2: Node = nodes_from_values(values=values1)
    node3: Node = nodes_from_values(values=values2)

    # Execution
    split(first_head=node0, second_head=node1)

    # Validation
    assert equal(first=node0, second=node2)
    assert equal(first=node1, second=node3)
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True


def test_split_three_nodes_two_heads_one_list_success():
    # Setup
    values0: list[str] = ["zero", "one", "two"]
    values1: list[str] = ["zero", "one"]
    values2: list[str] = ["two"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next.next
    node2: Node = nodes_from_values(values=values1)
    node3: Node = nodes_from_values(values=values2)

    # Execution
    split(first_head=node0, second_head=node1)

    # Validation
    assert equal(first=node0, second=node2)
    assert equal(first=node1, second=node3)
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True


def test_split_four_nodes_two_heads_one_list_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three"]
    values1: list[str] = ["zero", "one"]
    values2: list[str] = ["two", "three"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next.next
    node2: Node = nodes_from_values(values=values1)
    node3: Node = nodes_from_values(values=values2)

    # Execution
    split(first_head=node0, second_head=node1)

    # Validation
    assert equal(first=node0, second=node2)
    assert equal(first=node1, second=node3)
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True


def test_split_two_nodes_two_heads_two_lists_failure():
    # Setup
    values0: list[str] = ["zero", "one"]
    values1: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values1)

    # Validation
    with pytest.raises(NodesNotInSameListError):
        split(first_head=node0, second_head=node1)



from src.py_cdll import CDLL


def test_insert_with_empty_list_success():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL()
    index_requested: int = 2
    index_inserted: int = 0

    # Execution
    cdll0.insert(index=index_requested, value=data0)

    # Validation
    assert cdll0[index_inserted] == data0


def test_insert_after_last_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)
    index_requested: int = 2

    # Execution
    cdll0.insert(index=index_requested, value=data2)

    # Validation
    assert cdll0.last == data2


def test_insert_way_out_of_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)
    index_requested: int = 222
    index_inserted: int = 2

    # Execution
    cdll0.insert(index=index_requested, value=data2)

    # Validation
    assert cdll0[index_inserted] == data2


def test_insert_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0.insert(index=2, value=data4)

    # Validation
    assert cdll0[3] == data2


def test_insert_index_zero_in_empty_list_success():
    # Setup
    index0: int = 0
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Execution
    cdll0._insert(index=index0, value=data0)

    # Validation
    assert cdll0[0] == data0


def test_insert_index_seven_in_empty_list_success():
    # Setup
    index0: int = 7
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Execution
    cdll0._insert(index=index0, value=data0)

    # Validation
    assert cdll0[0] == data0


def test_insert_index_zero_update_head_next_prev_connections_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._insert(index=0, value=data5)

    # Validation
    assert cdll0.head == data5
    assert cdll0._head.next.value == data0
    assert cdll0._head.previous.value == data4
    assert cdll0._head.next.previous.value == data5
    assert cdll0._head.previous.next.value == data5



import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import Node


def test__node_remove_at_index_positive_index_in_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = 2

    # Execution
    node0: Node = cdll0._node_remove_at_index(index=index0)

    # Validation
    assert node0.value == data2
    assert data2 not in cdll0


def test__node_remove_at_index_negative_index_in_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = -2

    # Execution
    node0: Node = cdll0._node_remove_at_index(index=index0)

    # Validation
    assert node0.value == data3
    assert data3 not in cdll0


def test__node_remove_at_index_positive_index_out_of_range_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = 11

    # Validation
    with pytest.raises(IndexError):
        cdll0._node_remove_at_index(index=index0)


def test__node_remove_at_index_negative_index_out_of_range_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = -8

    # Validation
    with pytest.raises(IndexError):
        cdll0._node_remove_at_index(index=index0)



import pytest

from src.py_cdll import CDLL, EmptyCDLLError


def test_head_of_empty_list_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        _ = cdll0.head


def test_head_after_init_success():
    # Setup
    data: str = "data"
    cdll0 = CDLL(values=[data])

    # Validation
    assert cdll0.head == data
    assert cdll0.last == data



from .circular_doubly_linked_list import CDLL
from .exceptions import BaseCDLLException, CDLLAlreadyPopulatedError, NegativeIndexError, \
    ValueNotFoundError, FirstValueNotFoundError, SecondValueNotFoundError, ValuesNotAdjacentError, \
    NoAdjacentValueError, MultipleValuesFoundError, EmptyCDLLError, UnevenListLengthError, InputNotIterableError, \
    NoBeforeAndAfterUniqueError, AmbiguousEmptyCDLLStateError, NotANodeError, UnableToPopulateWithNoValuesError, \
    NodesNotInSameListError, NoNewHeadError



from typing import Iterator, Any

import pytest

from src.py_cdll import CDLL


def test___reversed___empty_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Execution
    iterator0: Iterator[Any] = reversed(cdll0)

    # Validation
    with pytest.raises(StopIteration):
        next(iterator0)


def test___reversed___one_yields_one_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[Any] = reversed(cdll0)

    # Validation
    assert len(list(iterator0)) == 1


def test___reversed___three_yields_three_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[Any] = reversed(cdll0)

    # Validation
    assert len(list(iterator0)) == 3


def test___reversed___three_yields_three_index_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[Any] = reversed(cdll0)

    # Validation
    assert next(iterator0) == data2
    assert next(iterator0) == data1
    assert next(iterator0) == data0



from src.py_cdll import CDLL


def test__slice_start_to_end_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data0)
    cdll1.append(data1)
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)

    slice0: slice = slice(None)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_inside_to_end_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)

    slice0: slice = slice(2, None)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_start_to_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data0)
    cdll1.append(data1)
    cdll1.append(data2)

    slice0: slice = slice(None, 3)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_inside_to_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data2)
    cdll1.append(data3)

    slice0: slice = slice(2, 4)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_outside_to_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(9, 3)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_inside_to_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)

    slice0: slice = slice(2, 9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_outside_to_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(7, 9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_inside_to_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data1)
    cdll1.append(data2)
    cdll1.append(data3)

    slice0: slice = slice(-4, 4)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_inside_to_negative_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data1)
    cdll1.append(data2)

    slice0: slice = slice(1, -2)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_inside_to_negative_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data2)
    cdll1.append(data3)

    slice0: slice = slice(-3, -1)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_outside_to_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data0)
    cdll1.append(data1)
    cdll1.append(data2)

    slice0: slice = slice(-9, 3)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_inside_to_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data1)
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)

    slice0: slice = slice(-4, 9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_outside_to_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data0)
    cdll1.append(data1)
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)

    slice0: slice = slice(-9, 9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_outside_to_negative_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(9, -4)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_inside_to_negative_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(2, -9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_outside_to_negative_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(9, -9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_outside_to_negative_inside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()
    cdll1.append(data0)
    cdll1.append(data1)

    slice0: slice = slice(-9, -3)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_inside_to_negative_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(-3, -9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2


def test__slice_negative_outside_to_negative_outside_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    cdll1: CDLL = CDLL()

    slice0: slice = slice(-6, -9)

    # Execution
    cdll2: CDLL = cdll0._slice(segment=slice0)

    # Validation
    assert cdll1 == cdll2



import pytest

from src.py_cdll import CDLL


def test__replace_at_index_empty_cdll_index_zero_failure():
    # Setup
    index0: int = 0
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0._replace_at_index(index=index0, replacement=data0)


def test__replace_at_index_empty_cdll_index_seven_failure():
    # Setup
    index0: int = 7
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0._replace_at_index(index=index0, replacement=data0)


def test__replace_at_index_length_three_cdll_index_zero_success():
    # Setup
    index0: int = 0
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._replace_at_index(index=index0, replacement=data0)

    # Validation
    assert cdll0[index0] == data0


def test__replace_at_index_length_three_cdll_index_six_failure():
    # Setup
    index0: int = 7
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(IndexError):
        cdll0._replace_at_index(index=index0, replacement=data0)


def test__replace_at_index_zero_with_empty_list_failure():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0[0] = data0


def test__replace_at_index_with_empty_list_failure():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0[2] = data0


def test__replace_at_index_greater_than_list_length_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(data1)

    # Validation
    with pytest.raises(IndexError):
        cdll0[2] = data0


def test__replace_at_index_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    data_new1: str = "inserted1"
    data_new2: str = "inserted2"
    cdll0[2] = data_new1
    cdll0[3] = data_new2

    # Validation
    assert cdll0[0] == data0
    assert cdll0[1] == data1
    assert cdll0[2] == data_new1
    assert cdll0[3] == data_new2
    assert len(cdll0) == 4


def test__replace_at_index_zero_updates_head_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data_new: str = "new_head"
    cdll0: CDLL = CDLL(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)

    # Execution
    cdll0[0] = data_new

    # Validation
    assert cdll0.head == data_new


def test__replace_at_index_zero_of_empty_list_failure():
    # Setup
    data: str = "data"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0[0] = data



from src.py_cdll import CDLL


def test__are_consecutive_integers_consecutive_success():
    # Setup
    integers0: list[int] = [1, 2, 3, 4, 5]
    consecutive0: bool = True

    # Execution
    consecutive1: bool = CDLL._are_consecutive(numbers=integers0)

    # Validation
    assert consecutive0 == consecutive1


def test__are_consecutive_integers_not_consecutive_success():
    # Setup
    integers0: list[int] = [1, 4, 5]
    consecutive0: bool = False

    # Execution
    consecutive1: bool = CDLL._are_consecutive(numbers=integers0)

    # Validation
    assert consecutive0 == consecutive1



import pytest

from src.py_cdll import NoNewHeadError
from src.py_cdll.node import Node, insert_between, nodes_from_values, equal, is_consistent


def test_insert_between_first_zero_second_one_success():
    # Setup
    head0: Node = nodes_from_values(values=[0])
    head1: Node = nodes_from_values(values=[1])
    head2: Node = nodes_from_values(values=[0, 1])
    before0: Node = head0
    after0: Node = head0.next
    insert0: Node = head1
    remove0: None = None

    # Execution
    removed0: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert removed0 is remove0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True


def test_insert_between_first_seven_second_none_success():
    # Setup
    head0: Node = nodes_from_values(values=[7])
    head1: None = None
    head2: Node = nodes_from_values(values=[7])
    before0: Node = head0
    after0: Node = head0.next
    insert0: None = head1

    # Execution
    inserted0: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert inserted0 is insert0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True


def test_insert_between_first_zero_two_second_one_success():
    # Setup
    head0: Node = nodes_from_values(values=[0, 2])
    head1: Node = nodes_from_values(values=[1])
    head2: Node = nodes_from_values(values=[0, 1, 2])
    before0: Node = head0
    after0: Node = head0.next
    insert0: Node = head1
    remove0: None = None

    # Execution
    removed0: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert removed0 is remove0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True


def test_insert_between_first_three_nine_second_none_success():
    # Setup
    head0: Node = nodes_from_values(values=[3, 9])
    head1: None = None
    head2: Node = nodes_from_values(values=[3, 9])
    before0: Node = head0
    after0: Node = head0.next
    insert0: None = head1

    # Execution
    inserted0: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert inserted0 is insert0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True


def test_insert_between_first_three_nine_second_none_remove_success():
    # Setup
    head0: Node = nodes_from_values(values=[3, 9])
    head1: None = None
    head2: Node = nodes_from_values(values=[3])
    before0: Node = head0
    after0: Node = head0
    insert0: None = head1
    remove0: Node = head0.next

    # Execution
    removed0: Node = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert removed0 is remove0
    assert equal(first=removed0, second=remove0)
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=removed0) is True


def test_insert_between_first_five_twelve_second_two_success():
    # Setup
    head0: Node = nodes_from_values(values=[5, 12])
    head1: Node = nodes_from_values(values=[2])
    head2: Node = nodes_from_values(values=[2, 5, 12])
    before0: Node = head0.previous
    after0: Node = head0
    insert0: Node = head1
    remove0: None = None

    # Execution
    removed0: None = insert_between(before=before0, after=after0, insert=insert0, head=insert0)

    # Validation
    assert removed0 is remove0
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True


def test_insert_between_first_five_twelve_second_two_four_success():
    # Setup
    head0: Node = nodes_from_values(values=[5, 12])
    head1: Node = nodes_from_values(values=[2, 4])
    head2: Node = nodes_from_values(values=[2, 4, 5, 12])
    before0: Node = head0.previous
    after0: Node = head0
    insert0: Node = head1
    remove0: None = None

    # Execution
    removed0: None = insert_between(before=before0, after=after0, insert=insert0, head=insert0)

    # Validation
    assert removed0 is remove0
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True


def test_insert_between_first_five_twelve_sixteen_twenty_second_none_success():
    # Setup
    head0: Node = nodes_from_values(values=[5, 12, 16, 20])
    head1: None = None
    head2: Node = nodes_from_values(values=[5, 20])
    before0: Node = head0
    after0: Node = head0.previous
    insert0: None = head1
    removed0: Node = nodes_from_values(values=[12, 16])

    # Execution
    inserted0: Node = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert equal(first=inserted0, second=removed0)
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=inserted0) is True


def test_insert_between_first_three_seven_nine_fifteen_nineteen_second_ten_twelve_two_five_without_head_failure():
    # Setup
    head0: Node = nodes_from_values(values=[3, 7, 9, 15, 19])
    head1: Node = nodes_from_values(values=[10, 12, 2, 5])
    before0: Node = head0.next.next
    after0: Node = before0
    insert0: Node = head1

    # Validation
    with pytest.raises(NoNewHeadError):
        _: Node = insert_between(before=before0, after=after0, insert=insert0)


def test_insert_between_first_three_seven_nine_fifteen_nineteen_second_ten_twelve_two_five_with_head_success():
    # Setup
    head0: Node = nodes_from_values(values=[3, 7, 9, 15, 19])
    head1: Node = nodes_from_values(values=[10, 12, 2, 5])
    head2: Node = nodes_from_values(values=[2, 5, 9, 10, 12])
    head3: Node = head1.previous.previous
    before0: Node = head0.next.next
    after0: Node = before0
    insert0: Node = head1
    removed0: Node = nodes_from_values(values=[15, 19, 3, 7])

    # Execution
    inserted0: Node = insert_between(before=before0, after=after0, insert=insert0, head=head3)

    # Validation
    assert equal(first=inserted0, second=removed0)
    assert equal(first=head3, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_consistent(node=inserted0) is True



import pytest

from src.py_cdll import CDLL


def test___radd___wrong_type_failure():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    adder0: int = 22

    # Validation
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        adder0 + cdll0


def test___radd___list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=[])
    cdll2: CDLL = CDLL(values=datas0 + datas0)

    # Execution
    cdll3: CDLL = datas0 + cdll0

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___radd___empty_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=[])
    cdll2: CDLL = CDLL(values=datas0)

    # Execution
    cdll3: CDLL = [] + cdll0

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___radd___one_value_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll3: CDLL = datas1 + cdll0

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___radd___two_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll3: CDLL = datas1 + cdll0

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___radd___five_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    data7: str = "data7"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4, data5, data6, data7]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll3: CDLL = datas1 + cdll0

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___radd___item_and_empty_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    datas1: list[str] = []
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll2: CDLL = datas1 + cdll0

    # Validation
    assert cdll2 == cdll1
    assert cdll2 is not cdll0 and cdll2 is not cdll1


def test___radd___empty_and_item_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = []
    datas1: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll2: CDLL = datas1 + cdll0
    # Validation
    assert cdll2 == cdll1
    assert cdll2 is not cdll0 and cdll2 is not cdll1


def test___radd___item_and_item_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0]
    datas1: list[str] = [data1]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll2: CDLL = datas1 + cdll0

    # Validation
    assert cdll2 == cdll1
    assert cdll2 is not cdll0 and cdll2 is not cdll1


def test___radd___two_items_and_five_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1]
    datas1: list[str] = [data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll2: CDLL = datas1 + cdll0

    # Validation
    assert cdll2 == cdll1
    assert cdll2 is not cdll0 and cdll2 is not cdll1


def test___radd___three_items_and_four_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll2: CDLL = datas1 + cdll0

    # Validation
    assert cdll2 == cdll1
    assert cdll2 is not cdll0 and cdll2 is not cdll1


def test___radd___cdll_tuple_success():
    # Setup
    cdll0: CDLL = CDLL()
    tuple1: tuple = tuple()

    # Execution
    cdll1: CDLL = tuple1 + cdll0

    # Validation
    assert cdll1 == cdll0
    assert cdll1 is not cdll0


def test___radd___cdll_list_success():
    # Setup
    cdll0: CDLL = CDLL()
    list1: list = list()

    # Execution
    cdll1: CDLL = list1 + cdll0

    # Validation
    assert cdll1 == cdll0
    assert cdll1 is not cdll0


def test___radd___cdll_set_success():
    # Setup
    cdll0: CDLL = CDLL()
    set1: set = set()

    # Execution
    cdll1: CDLL = set1 + cdll0

    # Validation
    assert cdll1 == cdll0
    assert cdll1 is not cdll0


def test___radd___cdll_range_success():
    # Setup
    cdll0: CDLL = CDLL(values=[5, 6, 7, 8, 9])
    cdll1: CDLL = CDLL(values=[0, 1, 2, 3, 4] + [5, 6, 7, 8, 9])
    range1: range = range(5)

    # Execution
    cdll2: CDLL = range1 + cdll0

    # Validation
    assert cdll2 == cdll1



import pytest

from src.py_cdll import CDLL, EmptyCDLLError
from src.py_cdll.node import Node


def test__node_at_index_first_index_in_zero_items_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        cdll0._node_at_index(index=1)


def test__node_at_index_eighth_index_in_four_items_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(IndexError):
        cdll0._node_at_index(index=8)


def test__node_at_index_zeroth_index_in_one_item_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    node0: Node = cdll0._node_at_index(index=0)

    # Validation
    assert node0 is cdll0._head


def test__node_at_index_first_index_in_two_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    node0: Node = cdll0._node_at_index(index=1)

    # Validation
    assert node0 is cdll0._last


def test__node_at_index_second_index_in_four_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    node0: Node = cdll0._node_at_index(index=2)

    # Validation
    assert node0 is cdll0._head.next.next
    assert node0 is cdll0._last.previous



import pytest

from src.py_cdll import CDLL, ValueNotFoundError, EmptyCDLLError
from src.py_cdll.compare import Comparison


def test__node_first_single_option_single_hit_identity_success():
    # Setup
    index0: int = 0
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index1, node0 = cdll0._node_first(value=data0, comparison=Comparison.IDENTITY)

    # Validation
    assert index0 == index1
    assert data0 is node0.value


def test__node_first_single_option_single_hit_equality_success():
    # Setup
    index0: int = 0
    data0: str = "data0"
    data1: str = "data0"
    datas0: list[str] = [data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index1, node0 = cdll0._node_first(value=data0, comparison=Comparison.EQUALITY)

    # Validation
    assert index0 == index1
    assert data0 == node0.value


def test__node_first_multiple_options_multiple_equal_single_hit_identity_success():
    # Setup
    index0: int = 2
    data0: list[str] = ["data0"]
    data1: list[str] = ["data0"]
    data2: list[str] = ["data0"]
    datas0: list[list[str]] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index1, node0 = cdll0._node_first(value=data2, comparison=Comparison.IDENTITY)

    # Validation
    assert index0 == index1
    assert data2 is node0.value


def test__node_first_multiple_options_single_hit_equality_success():
    # Setup
    index0: int = 3
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data0"
    datas0: list[str] = [data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index1, node0 = cdll0._node_first(value=data0, comparison=Comparison.EQUALITY)

    # Validation
    assert index0 == index1
    assert data0 == node0.value


def test__node_first_multiple_options_first_hit_identity_success():
    # Setup
    index0: int = 1
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data0"
    datas0: list[str] = [data1, data0, data2, data3, data0, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index1, node0 = cdll0._node_first(value=data0, comparison=Comparison.IDENTITY)

    # Validation
    assert index0 == index1
    assert data0 == node0.value


def test__node_first_multiple_options_first_hit_equality_failure():
    # Setup
    index0: int = 1
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data0"
    datas0: list[str] = [data1, data4, data2, data3, data4, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index1, node0 = cdll0._node_first(value=data0, comparison=Comparison.EQUALITY)

    # Validation
    assert index0 == index1
    assert data0 == node0.value


def test__node_first_no_options_no_hits_failure():
    # Setup
    data0: str = "data0"
    datas0: list = []
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(EmptyCDLLError):
        cdll0._node_first(value=data0, comparison=Comparison.IDENTITY)


def test__node_first_single_option_no_hits_identity_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list = [data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueNotFoundError):
        cdll0._node_first(value=data0, comparison=Comparison.IDENTITY)


def test__node_first_single_option_no_hits_equality_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list = [data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueNotFoundError):
        cdll0._node_first(value=data0, comparison=Comparison.EQUALITY)


def test__node_first_multiple_options_no_hits_identity_failure():
    # Setup
    data0: list[str] = ["data0"]
    data1: list[str] = ["data1"]
    data2: list[str] = ["data2"]
    data3: list[str] = ["data0"]
    datas0: list[list[str]] = [data1, data3, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueNotFoundError):
        cdll0._node_first(value=data0, comparison=Comparison.IDENTITY)


def test__node_first_multiple_options_no_hits_equality_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list = [data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueNotFoundError):
        cdll0._node_first(value=data0, comparison=Comparison.EQUALITY)



import pytest

from src.py_cdll import CDLL


def test__replace_with_slice_replace_single_empty_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 4, 5]
    datas2: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 3, 1)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_multiple_empty_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 5]
    datas2: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(1, 4, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_single_in_middle_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 333, 4, 5]
    datas2: list[int] = [333]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 3, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_multiple_in_middle_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 222, 23, 333, 34, 444, 5]
    datas2: list[int] = [222, 23, 333, 34, 444]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(1, 4, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_single_in_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 555]
    datas2: list[int] = [555]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(4, 5, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_multiple_from_tail_past_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 444, 555, 666, 777, 888]
    datas2: list[int] = [444, 555, 666, 777, 888]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(3, 5, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_multiple_spread_out_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [0, 1, 222, 3, 4, 555, 6, 7, 8, 9]
    datas2: list[int] = [222, 555]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 7, 3)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_replace_multiple_spread_out_with_too_few_replacement_values_failure():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas2: list[int] = [111, 222, 333]
    cdll0: CDLL[int] = CDLL(values=datas0)
    slice0: slice = slice(0, None, 2)

    # Validation
    with pytest.raises(ValueError):
        cdll0._replace_with_slice(segment=slice0, values=datas2)


def test__replace_with_slice_replace_multiple_spread_out_with_too_many_replacement_values_failure():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas2: list[int] = [111, 222, 333, 444, 555, 666]
    cdll0: CDLL[int] = CDLL(values=datas0)
    slice0: slice = slice(0, None, 2)

    # Validation
    with pytest.raises(ValueError):
        cdll0._replace_with_slice(segment=slice0, values=datas2)


def test__replace_with_slice_replace_all_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [111, 222, 333, 444, 555]
    datas2: list[int] = [111, 222, 333, 444, 555]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(0, None, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_add_nothing_to_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5]
    datas2: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(5, None, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_add_single_to_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5, 666]
    datas2: list[int] = [666]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(5, None, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_add_single_to_last_with_high_index_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5, 666]
    datas2: list[int] = [666]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(555, None, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1


def test__replace_with_slice_add_multiple_to_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5, 666, 777, 888, 999]
    datas2: list[int] = [666, 777, 888, 999]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(5, None, None)

    # Execution
    cdll0._replace_with_slice(segment=slice0, values=datas2)

    # Validation
    assert cdll0 == cdll1



from src.py_cdll import CDLL


def test__head_empty_success():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    assert cdll0._head is None


def test__head_three_items_success():
    # Setup
    data0: int = 0
    data1: int = 1
    data2: int = 2
    datas0: list[int] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    assert cdll0._head.previous is cdll0._head.next.next
    assert cdll0._head.previous.previous is cdll0._head.next
    assert cdll0._head.previous.previous.previous is cdll0._head
    assert cdll0._head.next.next.next is cdll0._head



from enum import Enum, auto
from typing import Any


class Comparison(Enum):
    EQUALITY = auto()
    IDENTITY = auto()
    IDENTITY_OR_EQUALITY = auto()


def compare_identity(pair: tuple[Any, Any]) -> bool:
    is_identical: bool = pair[0] is pair[1]
    return is_identical


def compare_equality(pair: tuple[Any, Any]) -> bool:
    is_equal: bool = pair[0] == pair[1]
    return is_equal


def compare(pair: tuple[Any, Any], comparison: Comparison) -> bool:
    is_comparable: bool = False

    match comparison:
        case Comparison.IDENTITY:
            is_comparable = compare_identity(pair=pair)
        case Comparison.EQUALITY:
            is_comparable = compare_equality(pair=pair)
        case Comparison.IDENTITY_OR_EQUALITY:
            is_comparable = compare_identity(pair=pair) or compare_equality(pair=pair)

    return is_comparable



from src.py_cdll.node import nodes_from_values, Node, is_consistent


def test_is_consistent_one_node_success():
    # Setup
    values0: list[str] = ["zero"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    is_alright: bool = is_consistent(node=node0)

    # Validation
    assert is_alright is True


def test_is_consistent_two_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    is_alright: bool = is_consistent(node=node0)

    # Validation
    assert is_alright is True


def test_is_consistent_seven_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    is_alright: bool = is_consistent(node=node0)

    # Validation
    assert is_alright is True


def test_is_consistent_eight_nodes_wrong_head_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six", "seven"]
    node0: Node = nodes_from_values(values=values0)
    is_alright0: bool = is_consistent(node=node0)

    # Verification
    assert is_alright0 is True

    # Execution
    node1: Node = Node(value=None)
    node0.next.next.next.next.next.head = node1
    is_alright1: bool = is_consistent(node=node0)

    # Validation
    assert is_alright1 is False


def test_is_consistent_nine_nodes_wrong_index_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six", "seven", "eight"]
    node0: Node = nodes_from_values(values=values0)
    is_alright0: bool = is_consistent(node=node0)

    # Verification
    assert is_alright0 is True

    # Execution
    index0: int = 999999
    node0.next.next.next.next.next.next.index = index0
    is_alright1: bool = is_consistent(node=node0)

    # Validation
    assert is_alright1 is False


def test_is_consistent_ten_nodes_wrong_previous_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    node0: Node = nodes_from_values(values=values0)
    is_alright0: bool = is_consistent(node=node0)

    # Verification
    assert is_alright0 is True

    # Execution
    node1: Node = Node(value=None)
    node0.next.next.next.next.next.next.next.previous = node1
    is_alright1: bool = is_consistent(node=node0)

    # Validation
    assert is_alright1 is False



from typing import Iterator

from src.py_cdll import CDLL
from src.py_cdll.compare import Comparison
from src.py_cdll.node import Node


def test__nodes_value_zero_yields_zero_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL()
    cdll0.append(value=data0)
    cdll0.append(value=data1)
    cdll0.append(value=data2)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value(value=data, comparison=Comparison.IDENTITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_one_yields_one_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL()
    cdll0.append(value=data0)
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll0.append(value=data)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value(value=data, comparison=Comparison.IDENTITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_three_yields_three_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL()
    cdll0.append(value=data)
    cdll0.append(value=data0)
    cdll0.append(value=data1)
    cdll0.append(value=data)
    cdll0.append(value=data2)
    cdll0.append(value=data)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value(value=data, comparison=Comparison.IDENTITY)

    # Validation
    assert len(list(iterator0)) == 3


def test__nodes_value_three_yields_three_indices_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL()
    cdll0.append(value=data)
    cdll0.append(value=data0)
    cdll0.append(value=data1)
    cdll0.append(value=data)
    cdll0.append(value=data2)
    cdll0.append(value=data)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value(value=data, comparison=Comparison.IDENTITY)

    # Validation
    assert next(iterator0)[0] == 0
    assert next(iterator0)[0] == 3
    assert next(iterator0)[0] == 5



import pytest

from src.py_cdll import CDLL
from src.py_cdll.exceptions import CDLLAlreadyPopulatedError, UnableToPopulateWithNoValuesError


def test__populate_non_empty_cdll_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0]
    datas1: list[str] = [data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(CDLLAlreadyPopulatedError):
        cdll0._populate(values=datas1)


def test__populate_empty_cdll_with_no_values_failure():
    # Setup
    datas0: list[str] = []
    datas1: list[str] = []
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(UnableToPopulateWithNoValuesError):
        cdll0._populate(values=datas1)


def test__populate_empty_cdll_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL()

    # Execution
    cdll0._populate(values=datas0)

    # Validation
    assert cdll0.head == data0
    assert cdll0._head == cdll0._head.next
    assert cdll0._head == cdll0._head.previous
    assert len(cdll0) == 1



import pytest

from src.py_cdll.circular_doubly_linked_list import CDLL, Connectivity
from src.py_cdll.exceptions import EmptyCDLLError, FirstValueNotFoundError, \
    SecondValueNotFoundError, ValuesNotAdjacentError


########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################


def test_shift_head_forwards_empty_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        cdll0._shift_head_forwards()


def test_shift_head_forwards_with_full_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._shift_head_forwards()

    # Validation
    assert cdll0.head == data1
    assert cdll0.last == data0


def test_shift_head_forwards_multiple_with_full_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._shift_head_forwards(5)

    # Validation
    assert cdll0.head == data2
    assert cdll0.last == data1


def test_shift_head_forwards_zero_amount_with_full_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0._shift_head_forwards(0)


def test_shift_head_forwards_negative_amount_with_full_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0._shift_head_forwards(-1)


########################################################################################################################


def test_shift_head_backwards_empty_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        cdll0._shift_head_backwards()


def test_shift_head_backwards_with_full_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._shift_head_backwards()

    # Validation
    assert cdll0.head == data2
    assert cdll0.last == data1


def test_shift_head_backwards_multiple_with_full_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._shift_head_backwards(5)

    # Validation
    assert cdll0.head == data1
    assert cdll0.last == data0


def test_shift_head_backwards_zero_amount_with_full_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0._shift_head_backwards(0)


def test_shift_head_backwards_negative_amount_with_full_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0._shift_head_backwards(-1)


########################################################################################################################


def test_rotate_none_success():
    # Setup
    data0: str = "0"
    data1: str = "1"
    data2: str = "2"
    data3: str = "3"
    data4: str = "4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas0)

    # Validation
    assert cdll0 == cdll1


def test_rotate_positive_one_success():
    # Setup
    data0: str = "0"
    data1: str = "1"
    data2: str = "2"
    data3: str = "3"
    data4: str = "4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)
    cdll1: CDLL = CDLL()
    cdll1.append(data1)
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)
    cdll1.append(data0)

    # Execution
    cdll1.rotate(amount=1)

    # Validation
    assert cdll0 == cdll1


def test_rotate_positive_two_success():
    # Setup
    data0: str = "0"
    data1: str = "1"
    data2: str = "2"
    data3: str = "3"
    data4: str = "4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)
    cdll1: CDLL = CDLL()
    cdll1.append(data2)
    cdll1.append(data3)
    cdll1.append(data4)
    cdll1.append(data0)
    cdll1.append(data1)

    # Execution
    cdll1.rotate(amount=2)

    # Validation
    assert cdll0 == cdll1


def test_rotate_negative_one_success():
    # Setup
    data0: str = "0"
    data1: str = "1"
    data2: str = "2"
    data3: str = "3"
    data4: str = "4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)
    cdll1: CDLL = CDLL()
    cdll1.append(data4)
    cdll1.append(data0)
    cdll1.append(data1)
    cdll1.append(data2)
    cdll1.append(data3)

    # Execution
    cdll1.rotate(amount=-1)

    # Validation
    assert cdll0 == cdll1


def test_rotate_negative_two_success():
    # Setup
    data0: str = "0"
    data1: str = "1"
    data2: str = "2"
    data3: str = "3"
    data4: str = "4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)
    cdll1: CDLL = CDLL()
    cdll1.append(data3)
    cdll1.append(data4)
    cdll1.append(data0)
    cdll1.append(data1)
    cdll1.append(data2)

    # Execution
    cdll1.rotate(amount=-2)

    # Validation
    assert cdll0 == cdll1


########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################


def test_adjacent_with_adjacent_correct_order_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Execution
    is_adjacent: bool = cdll0.adjacent(pair=(data2, data3))

    # Validation
    assert is_adjacent


def test_adjacent_with_adjacent_incorrect_order_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Validation
    assert cdll0.adjacent(pair=(data4, data3)) is True


def test_adjacent_with_non_adjacent_correct_order_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Validation
    assert cdll0.adjacent(pair=(data1, data4)) is False


def test_adjacent_with_non_adjacent_incorrect_order_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL()
    cdll0.append(data0)
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Validation
    assert cdll0.adjacent(pair=(data3, data0)) is False


########################################################################################################################


def test_adjacency_direction_first_second_next_success():
    # Setup
    cdll0: CDLL = CDLL()
    data0: int = 0
    data1: int = 1
    cdll0.extend(values=[data0, data1])

    # Execution
    connectivity0: Connectivity = cdll0.adjacency_direction(pair=(data0, data1))

    # Validation
    assert connectivity0 is Connectivity.ADJACENT_NEXT


def test_adjacency_direction_second_first_previous_success():
    # Setup
    cdll0: CDLL = CDLL()
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    cdll0.extend(values=[data0, data1, data2, data3])

    # Execution
    connectivity0: Connectivity = cdll0.adjacency_direction(pair=(data0, data3))

    # Validation
    assert connectivity0 is Connectivity.ADJACENT_PREVIOUS


def test_adjacency_direction_separate_first_second_failure():
    # Setup
    cdll0: CDLL = CDLL()
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    cdll0.extend(values=[data0, data1, data2, data3, data4])

    # Validation
    with pytest.raises(ValuesNotAdjacentError):
        cdll0.adjacency_direction(pair=(data1, data3))


def test_adjacency_direction_separate_second_first_failure():
    # Setup
    cdll0: CDLL = CDLL()
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    cdll0.extend(values=[data0, data1, data2, data3, data4])

    # Validation
    with pytest.raises(ValuesNotAdjacentError):
        cdll0.adjacency_direction(pair=(data4, data2))


def test_adjacency_direction_missing_first_failure():
    # Setup
    cdll0: CDLL = CDLL()
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    data_missing: str = "missing"
    cdll0.extend(values=[data0, data1, data2, data3, data4])

    # Validation
    with pytest.raises(FirstValueNotFoundError):
        # noinspection PyTypeChecker
        cdll0.adjacency_direction(pair=(data_missing, data2))


def test_adjacency_direction_missing_second_failure():
    # Setup
    cdll0: CDLL = CDLL()
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    data_missing: str = "missing"
    cdll0.extend(values=[data0, data1, data2, data3, data4])

    # Validation
    with pytest.raises(SecondValueNotFoundError):
        # noinspection PyTypeChecker
        cdll0.adjacency_direction(pair=(data2, data_missing))


########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################


def test_mirror_empty_at_head_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0.mirror()


def test_mirror_single_item_at_head_success():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL(values=[data0])

    # Execution
    cdll0.mirror()

    # Validation
    assert cdll0.head == data0


def test_mirror_five_items_at_head_success():
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"

    cdll0: CDLL = CDLL(values=[data0])

    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Execution
    cdll0.mirror()

    # Validation
    assert list(cdll0) == [data0, data4, data3, data2, data1]


def test_mirror_six_items_at_index_two_success():
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"

    cdll0: CDLL = CDLL(values=[data0])

    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)
    cdll0.append(data5)

    index0: int = 2

    # Execution
    cdll0.mirror(index=index0)

    # Validation
    assert list(cdll0) == [data4, data3, data2, data1, data0, data5]


def test_mirror_six_items_at_index_minus_four_success():
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"

    cdll0: CDLL = CDLL(values=[data0])

    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)
    cdll0.append(data5)

    index0: int = -4

    # Execution
    cdll0.mirror(index=index0)

    # Validation
    assert list(cdll0) == [data4, data3, data2, data1, data0, data5]


########################################################################################################################


def test_switch_with_empty_index_out_of_bounds_success():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        cdll0.switch(pair=(0, 0))


def test_switch_with_index_out_of_bounds_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(data1)

    # Validation
    with pytest.raises(IndexError):
        cdll0.switch(pair=(0, 2))


def test_switch_with_single_item_success():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL(values=[data0])

    # Execution
    cdll0.switch(pair=(0, 0))

    # Validation
    assert cdll0[0] == data0


def test_switch_with_two_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(data1)

    # Execution
    cdll0.switch(pair=(0, 1))

    # Validation
    assert list(cdll0) == [data1, data0]


def test_switch_with_five_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Execution
    cdll0.switch(pair=(1, 3))

    # Validation
    assert list(cdll0) == [data0, data3, data2, data1, data4]


def test_switch_twice_with_five_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(data1)
    cdll0.append(data2)
    cdll0.append(data3)
    cdll0.append(data4)

    # Execution
    cdll0.switch(pair=(1, 3))
    cdll0.switch(pair=(2, 3))

    # Validation
    assert list(cdll0) == [data0, data3, data1, data2, data4]


########################################################################################################################


def test_equality_rotated_mirrored_empty_success():
    # Setup
    cdll0: CDLL = CDLL()
    cdll1: CDLL = CDLL()

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_one_element_success():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL(values=data0)
    cdll1: CDLL = CDLL(values=data0)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_two_elements_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=data0)
    cdll0.append(value=data1)
    cdll1: CDLL = CDLL(values=data0)
    cdll1.append(value=data1)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_two_by_two_different_elements_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    cdll0: CDLL = CDLL(values=data0)
    cdll0.append(value=data1)
    cdll1: CDLL = CDLL(values=data2)
    cdll1.append(value=data3)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert not is_equal


def test_equality_rotated_mirrored_five_elements_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL(values=data0)
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll0.append(value=data3)
    cdll0.append(value=data4)
    cdll1: CDLL = CDLL(values=data0)
    cdll1.append(value=data1)
    cdll1.append(value=data2)
    cdll1.append(value=data3)
    cdll1.append(value=data4)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_two_elements_shifted_one_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(value=data1)
    cdll1: CDLL = CDLL(values=[data1])
    cdll1.append(value=data0)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_five_elements_shifted_three_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll0.append(value=data3)
    cdll0.append(value=data4)
    cdll1: CDLL = CDLL(values=[data3])
    cdll1.append(value=data4)
    cdll1.append(value=data0)
    cdll1.append(value=data1)
    cdll1.append(value=data2)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_seven_elements_shifted_four_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll0.append(value=data3)
    cdll0.append(value=data4)
    cdll0.append(value=data5)
    cdll0.append(value=data6)
    cdll1: CDLL = CDLL(values=[data4])
    cdll1.append(value=data5)
    cdll1.append(value=data6)
    cdll1.append(value=data0)
    cdll1.append(value=data1)
    cdll1.append(value=data2)
    cdll1.append(value=data3)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_three_elements_shifted_two_and_mirrored_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll1: CDLL = CDLL(values=[data2])
    cdll1.append(value=data1)
    cdll1.append(value=data0)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_five_elements_shifted_three_and_mirrored_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll0.append(value=data3)
    cdll0.append(value=data4)
    cdll1: CDLL = CDLL(values=[data3])
    cdll1.append(value=data2)
    cdll1.append(value=data1)
    cdll1.append(value=data0)
    cdll1.append(value=data4)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


def test_equality_rotated_mirrored_seven_elements_shifted_five_and_mirrored_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    cdll0: CDLL = CDLL(values=[data0])
    cdll0.append(value=data1)
    cdll0.append(value=data2)
    cdll0.append(value=data3)
    cdll0.append(value=data4)
    cdll0.append(value=data5)
    cdll0.append(value=data6)
    cdll1: CDLL = CDLL(values=[data5])
    cdll1.append(value=data6)
    cdll1.append(value=data0)
    cdll1.append(value=data1)
    cdll1.append(value=data2)
    cdll1.append(value=data3)
    cdll1.append(value=data4)

    # Execution
    is_equal: bool = cdll0._eq_rotated_mirrored(other=cdll1)

    # Validation
    assert is_equal


########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################


if __name__ == '__main__':
    pass



from src.py_cdll import CDLL
from src.py_cdll.node import Node


def test__node_remove_from_single_success():
    # Setup
    list0: list[int] = [0]
    cdll0: CDLL = CDLL(values=list0)
    node0: Node = cdll0._head

    # Execution
    cdll0._node_remove(target=node0)

    # Validation
    assert len(cdll0) == 0


def test__node_remove_head_from_triple_success():
    # Bug: _node_remove had no check for whether head was being removed, which would invalidate pointer in CDLL
    # and cause infinite loops when trying to find head after that.

    # Setup
    list0: list[int] = [0, 1, 2]
    cdll0: CDLL = CDLL(values=list0)
    node0: Node = cdll0._head

    # Execution
    cdll0._node_remove(target=node0)

    # Validation
    assert len(cdll0) == 2


def test__node_remove_from_multiple_success():
    # Setup
    list0: list[int] = [0, 1, 2, 3, 4]
    cdll0: CDLL = CDLL(values=list0)
    node0: Node = cdll0._head.previous.previous

    # Execution
    cdll0._node_remove(target=node0)

    # Validation
    assert node0 not in cdll0._nodes()
    assert len(cdll0) == 4



from typing import Iterator

from src.py_cdll import CDLL
from src.py_cdll.compare import Comparison
from src.py_cdll.node import Node


def test__nodes_value_constrained_zero_yields_zero_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_one_yields_one_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2, data]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_constrained_three_yields_three_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data, data0, data1, data, data2, data]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 3


def test__nodes_value_constrained_three_yields_three_indices_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL()
    cdll0.append(value=data)
    cdll0.append(value=data0)
    cdll0.append(value=data1)
    cdll0.append(value=data)
    cdll0.append(value=data2)
    cdll0.append(value=data)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert next(iterator0)[0] == 0
    assert next(iterator0)[0] == 3
    assert next(iterator0)[0] == 5


def test__nodes_value_constrained_found_value_stop_out_of_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data1,
                                                                           stop=len(cdll0) + 1,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_constrained_not_found_value_stop_out_of_range_success():
    # Setup
    value0: str = "data5"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=value0,
                                                                           stop=len(cdll0) + 5,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_found_value_in_range_stop_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data2,
                                                                           stop=len(cdll0) // 2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_constrained_not_found_value_out_of_range_stop_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data4,
                                                                           stop=len(cdll0) // 2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_not_found_value_unavailable_stop_mid_list_success():
    # Setup
    value0: str = "data9"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=value0,
                                                                           stop=len(cdll0) // 2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_found_value_in_range_start_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data5,
                                                                           start=len(cdll0) // 2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_constrained_not_found_value_out_of_range_start_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data2,
                                                                           start=len(cdll0) // 2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_not_found_value_unavailable_start_mid_list_success():
    # Setup
    value0: str = "data9"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=value0,
                                                                           start=len(cdll0) // 2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_not_found_value_out_of_range_slice_zero_length_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data1,
                                                                           start=1,
                                                                           stop=1,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_not_found_value_unavailable_slice_zero_length_success():
    # Setup
    value0: str = "data6"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=value0,
                                                                           start=2,
                                                                           stop=2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_value_constrained_value_found_slice_start_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data2,
                                                                           start=-3,
                                                                           stop=3,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_constrained_value_found_slice_stop_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data2,
                                                                           start=2,
                                                                           stop=-2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_value_constrained_value_found_slice_both_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_value_constrained(value=data2,
                                                                           start=-3,
                                                                           stop=-2,
                                                                           comparison=Comparison.IDENTITY_OR_EQUALITY)

    # Validation
    assert len(list(iterator0)) == 1



from src.py_cdll import CDLL


def test_append_to_end_of_empty_list_success():
    # Setup
    data: str = "data"
    cdll0: CDLL = CDLL()

    # Execution
    cdll0.append(value=data)

    # Validation
    assert cdll0.head == data
    assert cdll0.last == data


def test_append_to_end_of_list_success():
    # Setup
    head: str = "head"
    data: str = "data"
    cdll0: CDLL = CDLL(values=[head])

    # Execution
    cdll0.append(data)

    # Validation
    assert cdll0.head == head
    assert cdll0.last == data


def test_append_retrieval_success():
    # Setup
    head: str = "head"
    data1: str = "data1"
    data2: str = "data2"
    cdll0: CDLL = CDLL(values=[head])

    # Execution
    cdll0.append(data1)
    cdll0.append(data2)

    # Validation
    assert cdll0.head == head
    assert cdll0.last == data2
    assert cdll0[0] == head
    assert cdll0[1] == data1
    assert cdll0[2] == data2



import pytest

from src.py_cdll import CDLL


def test___pop___empty_cdll_failure():
    # Setup
    index: int = 0
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0.pop(index=index)


def test___pop___index_outside_multiple_elements_range_failure():
    # Setup
    index: int = 9
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(IndexError):
        cdll0.pop(index=index)


def test___pop___index_inside_single_elements_success():
    # Setup
    index: int = 0
    data: str = "data"
    data0: str = "data"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    data1: str = cdll0.pop(index=index)

    # Validation
    assert data == data1


def test___pop___default_index_success():
    # Setup
    data0: str = "data"
    data1: str = "data"
    data2: str = "data"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    data3: str = cdll0.pop()

    # Validation
    assert data3 == data2


def test___pop___index_inside_multiple_elements_success():
    # Setup
    index: int = 3
    data: str = "data3"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    data1: str = cdll0.pop(index=index)

    # Validation
    assert data == data1



from src.py_cdll import CDLL
from src.py_cdll.node import Node, before_target, is_consistent


def test_before_first_zero_second_one_success():
    # Setup
    first: Node = CDLL(values=[0])._head
    second: Node = CDLL(values=[1])._head
    current: Node = first
    before_insert0: Node = first

    # Execution
    before_insert1: Node = before_target(current=current, head=first, target=second)

    # Validation
    assert before_insert0 == before_insert1
    assert is_consistent(node=before_insert1) is True


def test_before_first_one_second_one_success():
    # Setup
    first: Node = CDLL(values=[1])._head
    second: Node = CDLL(values=[1])._head
    current: Node = first
    before_insert0: Node = first

    # Execution
    before_insert1: Node = before_target(current=current, head=first, target=second)

    # Validation
    assert before_insert0 == before_insert1
    assert is_consistent(node=before_insert1) is True


def test_before_first_one_three_second_six_success():
    # Setup
    first: Node = CDLL(values=[1, 3])._head
    second: Node = CDLL(values=[6])._head
    current: Node = first
    before_insert0: Node = first.next

    # Execution
    before_insert1: Node = before_target(current=current, head=first, target=second)

    # Validation
    assert before_insert0 == before_insert1
    assert is_consistent(node=before_insert1) is True


def test_before_first_two_seven_second_four_six_success():
    # Setup
    first: Node = CDLL(values=[2, 7])._head
    second: Node = CDLL(values=[4, 6])._head
    current: Node = first
    before_insert0: Node = first

    # Execution
    before_insert1: Node = before_target(current=current, head=first, target=second)

    # Validation
    assert before_insert0 == before_insert1
    assert is_consistent(node=before_insert1) is True



from src.py_cdll.node import Node, equal, nodes_from_values


def test_equal_one_value_same_success():
    # Setup
    values0: list[str] = ["value0"]
    first0: Node = nodes_from_values(values=values0)
    second0: Node = nodes_from_values(values=values0)
    boolean0: bool = True

    # Execution
    boolean1: bool = equal(first=first0, second=second0)

    # Validation
    assert boolean0 is boolean1


def test_equal_one_value_different_values_success():
    # Setup
    values0: list[str] = ["value0"]
    values1: list[str] = ["value1"]
    first0: Node = nodes_from_values(values=values0)
    second0: Node = nodes_from_values(values=values1)
    boolean0: bool = False

    # Execution
    boolean1: bool = equal(first=first0, second=second0)

    # Validation
    assert boolean0 is boolean1


def test_equal_two_values_different_lengths_success():
    # Setup
    values0: list[str] = ["value0"]
    values1: list[str] = ["value0", "value1"]
    first0: Node = nodes_from_values(values=values0)
    second0: Node = nodes_from_values(values=values1)
    boolean0: bool = False

    # Execution
    boolean1: bool = equal(first=first0, second=second0)

    # Validation
    assert boolean0 is boolean1


def test_equal_two_values_same_success():
    # Setup
    values0: list[str] = ["value0", "value1"]
    first0: Node = nodes_from_values(values=values0)
    second0: Node = nodes_from_values(values=values0)
    boolean0: bool = True

    # Execution
    boolean1: bool = equal(first=first0, second=second0)

    # Validation
    assert boolean0 is boolean1


def test_equal_two_values_different_values_success():
    # Setup
    values0: list[str] = ["value0", "value1"]
    values1: list[str] = ["value0", "value2"]
    first0: Node = nodes_from_values(values=values0)
    second0: Node = nodes_from_values(values=values1)
    boolean0: bool = False

    # Execution
    boolean1: bool = equal(first=first0, second=second0)

    # Validation
    assert boolean0 is boolean1


def test_equal_two_values_same_index_head_different_success():
    # Setup
    values0: list[str] = ["value0", "value1"]
    values1: list[str] = ["value1", "value0"]
    first0: Node = nodes_from_values(values=values0)
    second0: Node = nodes_from_values(values=values1)
    second1: Node = second0.next
    boolean0: bool = False

    # Execution
    boolean1: bool = equal(first=first0, second=second1)

    # Validation
    assert boolean0 is boolean1



from src.py_cdll.node import Node, split_head_from_tail, nodes_from_values, is_consistent


def test_split_head_from_tail_one_node_success():
    # Setup
    values0: list[str] = ["zero"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    head0, tail0 = split_head_from_tail(node=node0)

    # Validation
    assert head0 is node0
    assert head0.next is node0
    assert head0.previous is node0
    assert tail0 is None
    assert is_consistent(node=head0) is True



def test_split_head_from_tail_two_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next

    # Execution
    head0, tail0 = split_head_from_tail(node=node0)

    # Validation
    assert head0 is node0
    assert head0.next is node0
    assert head0.previous is node0
    assert tail0 is node1
    assert tail0.next is node1
    assert tail0.previous is node1
    assert is_consistent(node=head0) is True
    assert is_consistent(node=tail0) is True


def test_split_head_from_tail_three_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one", "two"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next
    node2: Node = node0.next.next

    # Execution
    head0, tail0 = split_head_from_tail(node=node0)

    # Validation
    assert head0 is node0
    assert head0.next is head0
    assert head0.previous is head0
    assert tail0 is node1
    assert tail0.next is node2
    assert tail0.previous is node2
    assert is_consistent(node=head0) is True
    assert is_consistent(node=tail0) is True


def test_split_head_from_tail_four_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next
    node2: Node = node0.next.next
    node3: Node = node0.next.next.next

    # Execution
    head0, tail0 = split_head_from_tail(node=node0)

    # Validation
    assert head0 is node0
    assert head0.next is head0
    assert head0.previous is head0
    assert tail0 is node1
    assert tail0.next is node2
    assert tail0.previous is node3
    assert is_consistent(node=head0) is True
    assert is_consistent(node=tail0) is True



from src.py_cdll.node import Node, nodes_from_values, is_consistent


def test_nodes_from_values_empty_failure():
    # Setup
    datas0: list[str] = []
    values0: None = None

    # Execution
    values1: None = nodes_from_values(values=datas0)

    # Validation
    assert values1 == values0


def test_nodes_from_values_one_value_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]

    # Execution
    head0: Node = nodes_from_values(values=datas0)

    # Validation
    assert head0.value is data0
    assert head0.next is head0
    assert head0.previous is head0
    assert is_consistent(node=head0) is True


def test_nodes_from_values_two_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0, data1]

    # Execution
    head0: Node = nodes_from_values(values=datas0)

    # Validation
    assert head0.value is data0
    assert head0.next.value is data1
    assert head0.previous.value is data1
    assert head0.next.next is head0
    assert head0.previous.previous is head0
    assert is_consistent(node=head0) is True


def test_nodes_from_values_seven_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]

    # Execution
    head0: Node = nodes_from_values(values=datas0)

    # Validation
    assert head0.value is data0
    assert head0.next.value is data1
    assert head0.next.next.value is data2
    assert head0.next.next.next.value is data3
    assert head0.next.next.next.next.value is data4
    assert head0.next.next.next.next.next.value is data5
    assert head0.next.next.next.next.next.next.value is data6
    assert head0.next.next.next.next.next.next.next.value is data0
    assert head0.previous.value is data6
    assert head0.previous.previous.value is data5
    assert head0.previous.previous.previous.value is data4
    assert head0.previous.previous.previous.previous.value is data3
    assert head0.previous.previous.previous.previous.previous.value is data2
    assert head0.previous.previous.previous.previous.previous.previous.value is data1
    assert head0.previous.previous.previous.previous.previous.previous.previous.value is data0
    assert head0.next.next.next.next.next.next.next is head0
    assert head0.previous.previous.previous.previous.previous.previous.previous is head0
    assert is_consistent(node=head0) is True



from src.py_cdll.node import Node, middle_adjacent


def test_middle_adjacent_one_node_success():
    # Setup
    node0: Node = Node(value=None)
    node0.next = node0
    node0.previous = node0

    # Execution
    before_middle, after_middle = middle_adjacent(head=node0)

    # Validation
    assert before_middle is node0
    assert after_middle is node0


def test_middle_adjacent_two_nodes_success():
    # Setup
    node0: Node = Node(value=None)
    node1: Node = Node(value=None)
    node0.next = node1
    node1.previous = node0
    node1.next = node0
    node0.previous = node1

    # Execution
    before_middle, after_middle = middle_adjacent(head=node0)

    # Validation
    assert before_middle is node0
    assert after_middle is node1


def test_middle_adjacent_three_nodes_success():
    # Setup
    node0: Node = Node(value=None)
    node1: Node = Node(value=None)
    node2: Node = Node(value=None)
    node0.next = node1
    node1.previous = node0
    node1.next = node2
    node2.previous = node1
    node2.next = node0
    node0.previous = node2

    # Execution
    before_middle, after_middle = middle_adjacent(head=node0)

    # Validation
    assert before_middle is node1
    assert after_middle is node2


def test_middle_adjacent_four_nodes_success():
    # Setup
    node0: Node = Node(value=None)
    node1: Node = Node(value=None)
    node2: Node = Node(value=None)
    node3: Node = Node(value=None)
    node0.next = node1
    node1.previous = node0
    node1.next = node2
    node2.previous = node1
    node2.next = node3
    node3.previous = node2
    node3.next = node0
    node0.previous = node3

    # Execution
    before_middle, after_middle = middle_adjacent(head=node0)

    # Validation
    assert before_middle is node1
    assert after_middle is node2



from typing import Iterator

import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import Node


def test__nodes_empty_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes()

    # Validation
    with pytest.raises(StopIteration):
        next(iterator0)


def test__nodes_one_yields_one_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes()

    # Validation
    assert len(list(iterator0)) == 1


def test__nodes_three_yields_three_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes()

    # Validation
    assert len(list(iterator0)) == 3


def test__nodes_three_yields_three_index_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes()

    # Validation
    assert next(iterator0)[1].value == data0
    assert next(iterator0)[1].value == data1
    assert next(iterator0)[1].value == data2



from src.py_cdll import CDLL


def test___len___empty_list_success():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    assert len(cdll0) == 0


def test___len___with_one_item_success():
    # Setup
    cdll0: CDLL = CDLL(values=["data"])

    # Validation
    assert len(cdll0) == 1


def test___len___with_six_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    assert len(cdll0) == 6


def test___len___with_111_items_success():
    # Setup
    amount: int = 111
    data: range = range(amount)
    cdll0: CDLL = CDLL(values=data)

    # Validation
    assert len(cdll0) == amount



import pytest

from src.py_cdll import CDLL, AmbiguousEmptyCDLLStateError


def test_is_cdll_empty_zero_elements_success():
    # Setup
    cdll0: CDLL = CDLL()
    is_empty0: bool = True

    # Execution
    is_empty1: bool = cdll0.is_empty()

    # Validation
    assert is_empty1 is is_empty0


def test_is_cdll_empty_five_elements_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4]
    cdll0: CDLL = CDLL(values=datas0)
    is_empty0: bool = False

    # Execution
    is_empty1: bool = cdll0.is_empty()

    # Validation
    assert is_empty1 is is_empty0


def test_is_cdll_empty_five_elements_zero_length_failure():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4]
    cdll0: CDLL = CDLL(values=datas0)
    cdll0._length = 0

    # Validation
    with pytest.raises(AmbiguousEmptyCDLLStateError):
        cdll0.is_empty()


def test_is_cdll_empty_zero_elements_one_length_failure():
    # Setup
    cdll0: CDLL = CDLL()
    cdll0._length = 9

    # Validation
    with pytest.raises(AmbiguousEmptyCDLLStateError):
        cdll0.is_empty()



import pytest

from src.py_cdll import CDLL, NotANodeError
from src.py_cdll.node import Node


def test__insert_after_node_one_in_list_insert_not_node_failure():
    # Setup
    node0: int = 37
    cdll0: CDLL = CDLL(values=[None])

    # Validation
    with pytest.raises(NotANodeError):
        # noinspection PyTypeChecker
        cdll0._insert_after_node(anchor=cdll0._head, insert=node0)


def test__insert_after_node_one_in_list_insert_one_in_second_place_success():
    # Setup
    node0: Node = Node(value=1)
    data0: int = 0
    datas0: list[int] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._insert_after_node(anchor=cdll0._head, insert=node0)

    # Validation
    assert cdll0._head.next is node0
    assert cdll0._head.previous is node0
    assert node0.next is cdll0._head
    assert node0.previous is cdll0._head
    assert len(cdll0) == 2


def test__insert_after_node_two_in_list_insert_one_at_end_success():
    # Setup
    node0: Node = Node(value=2)
    data0: int = 0
    data1: int = 1
    datas0: list[int] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._insert_after_node(anchor=cdll0._last, insert=node0)

    # Validation
    assert cdll0._last is node0
    assert cdll0._head.next.next is node0
    assert cdll0._head.previous is node0
    assert node0.next is cdll0._head
    assert node0.previous.previous is cdll0._head
    assert len(cdll0) == 3


def test__insert_after_node_four_in_list_insert_one_in_middle_success():
    # Setup
    node0: Node = Node(value=2)
    data0: int = 0
    data1: int = 1
    data3: int = 3
    data4: int = 4
    datas0: list[int] = [data0, data1, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._insert_after_node(anchor=cdll0._head.next, insert=node0)

    # Validation
    assert cdll0._head.next.next is node0
    assert cdll0._last.previous.previous is node0
    assert node0.next.next is cdll0._last
    assert node0.previous.previous is cdll0._head
    assert len(cdll0) == 5



import pytest

from src.py_cdll import CDLL, EmptyCDLLError


def test___getitem___index_in_range_success():
    # Setup
    head: str = "head"
    cdll0: CDLL = CDLL(values=[head])

    # Execution
    first_item: str = cdll0[0]

    # Validation
    assert first_item is head


def test___getitem___index_out_of_range_failure():
    # Setup
    head: str = "head"
    cdll0: CDLL = CDLL(values=[head])

    # Validation
    with pytest.raises(IndexError):
        _ = cdll0[1]


def test___getitem___index_of_empty_list_out_of_range_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        _ = cdll0[0]


def test___getitem___negative_index_success():
    # Setup
    data0: int = 100
    data1: int = 200
    data2: int = 300
    datas0: list[int] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    last_item: int = cdll0[-1]

    # Validation
    assert last_item is data2
